/**
Copyright (c) 2000 Brian Nenninger

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**/

#include "hypergrid.h"
#include <stdio.h>
#include <stdlib.h>

hypergrid* hypergrid_create(int dim, int *sizes) {
  hypergrid *gridptr = (hypergrid *)malloc(sizeof(hypergrid));
  int i;
  int totalcells = 1;
  gridptr->num_dimensions = dim;
  gridptr->dimension_sizes = (int *)malloc(dim*sizeof(int));
  for(i=0; i<dim; i++) {
    gridptr->dimension_sizes[i] = sizes[i];
    totalcells *= sizes[i];
  }
  gridptr->num_cells = totalcells;
  gridptr->grid_data = (gridvalue_t *)calloc(totalcells, sizeof(gridvalue_t));
  return gridptr;
}

hypergrid* hypergrid_copy(hypergrid *srcgrid) {
  hypergrid *copy = hypergrid_create(srcgrid->num_dimensions, srcgrid->dimension_sizes);
  memcpy(copy->grid_data, srcgrid->grid_data, srcgrid->num_cells*sizeof(gridvalue_t));
  return copy;
}

void hypergrid_free(hypergrid *gridptr) {
  free(gridptr->dimension_sizes);
  free(gridptr->grid_data);
  free(gridptr);
}

int* hypergrid_coords_create(hypergrid *gridptr) {
  int *array = (int *)calloc(gridptr->num_dimensions, sizeof(int));
  return array;
}

void hypergrid_coords_free(int *coords) {
  free(coords);
}

int hypergrid_coords_isvalid(hypergrid *gridptr, int *coords) {
  int dim = gridptr->num_dimensions;
  int i;
  for(i=0; i<dim; i++) {
    if (coords[i]>=gridptr->dimension_sizes[i]) return 0;
  }
  return 1;
}

int hypergrid_coords_increment(hypergrid *gridptr, int *coords) {
  /* indices are in pseudo-big-endian order, so start incrementing from the last position */
  int ndim = gridptr->num_dimensions;
  int cdim = ndim;
  while (cdim--) {
    ++coords[cdim];
    if (coords[cdim]<gridptr->dimension_sizes[cdim]) {
      /* no overflow, we're done */
      return 1;
    }
    else {
      /* "carry" to next position */
      coords[cdim] = 0;
    }
  }
  /* either coords was the "last" one in the hypergrid or it was invalid */
  return 0;
}

/* "private" method to get the coords into the hypergrid's big 1-d array */
int _compute_coords_index(hypergrid *gridptr, int *coords) {
  int acc = 0;
  int placevalue = 1;
  int i=gridptr->num_dimensions;
  while (i--) {
    acc += (coords[i]*placevalue);
    placevalue *= gridptr->dimension_sizes[i];
  }
  return acc;
}

gridvalue_t hypergrid_get_value(hypergrid *gridptr, int *coords) {
  int index = _compute_coords_index(gridptr, coords);
  return gridptr->grid_data[index];
}

void hypergrid_set_value(hypergrid *gridptr, int *coords, gridvalue_t value) {
  int index = _compute_coords_index(gridptr, coords);
  gridptr->grid_data[index] = value;
}

/* "private" method used by hypergrid_neighbor_count */
int _count_neighbors(hypergrid *gridptr, int cmpequal, gridvalue_t cmpval, 
                     int baseindex, int dimnum, int placevalue, int wrap) {
  int next_placevalue = placevalue*gridptr->dimension_sizes[dimnum];
  int coord = baseindex / next_placevalue;
  int newindex;
  int count = 0;
  if (dimnum>0) {
    /* do three recursions corresponding to offsetting coordinate by -1, 0, +1 */
    int nextdim = dimnum-1;
    /* check 0 */
    newindex = baseindex;
    count += _count_neighbors(gridptr, cmpequal, cmpval, newindex, nextdim, next_placevalue, wrap);
    /* check -1 */
    newindex = baseindex-placevalue;
    /* if it wraps and the wrap flag is set, adjust index and check */
    if (newindex<0 || (newindex/next_placevalue!=coord)) {
      if (wrap) {
        newindex += next_placevalue;
        count += _count_neighbors(gridptr, cmpequal, cmpval, newindex, nextdim, next_placevalue, wrap);
      }
    }
    else {
      /* in range, always check */
      count += _count_neighbors(gridptr, cmpequal, cmpval, newindex, nextdim, next_placevalue, wrap);
    }
    /* check +1 */
    newindex = baseindex+placevalue;
    /* if it wraps and the wrap flag is set, adjust index and check */
    if ((newindex/next_placevalue!=coord)) {
      if (wrap) {
        newindex -= next_placevalue;
        count += _count_neighbors(gridptr, cmpequal, cmpval, newindex, nextdim, next_placevalue, wrap);
      }
    }
    else {
      /* in range, always check */
      count += _count_neighbors(gridptr, cmpequal, cmpval, newindex, nextdim, next_placevalue, wrap);
    }
    return count;
  }
  else {
    /* if we've hit dimension 0, return value directly */
    /* check 0 */
    gridvalue_t *griddata = gridptr->grid_data;
    newindex = baseindex;
    if ((cmpequal && griddata[newindex]==cmpval) || (!cmpequal && griddata[newindex]!=cmpval)) ++count;
    /* check -1 */
    newindex = baseindex-placevalue;
    /* if it wraps and the wrap flag is set, adjust index and check */
    if (newindex<0 || (newindex/next_placevalue!=coord)) {
      if (wrap) {
        newindex += next_placevalue;
        if ((cmpequal && griddata[newindex]==cmpval) || (!cmpequal && griddata[newindex]!=cmpval)) ++count;
      }
    }
    else {
      /* in range, always check */
      if ((cmpequal && griddata[newindex]==cmpval) || (!cmpequal && griddata[newindex]!=cmpval)) ++count;
    }
    /* check +1 */
    newindex = baseindex+placevalue;
    /* if it wraps and the wrap flag is set, adjust index and check */
    if ((newindex/next_placevalue!=coord)) {
      if (wrap) {
        newindex -= next_placevalue;
        if ((cmpequal && griddata[newindex]==cmpval) || (!cmpequal && griddata[newindex]!=cmpval)) ++count;
      }
    }
    else {
      /* in range, always check */
      if ((cmpequal && griddata[newindex]==cmpval) || (!cmpequal && griddata[newindex]!=cmpval)) ++count;
    }
    return count;
  }
  
}

/* cmpequal: true to count cells equal to cmpval, false to count cells not equal */
/* wrap: true to wrap around edges of grid */
int hypergrid_neighbor_count(hypergrid *gridptr, int *coords, int cmpequal, gridvalue_t cmpval, int wrap) {
  int index = _compute_coords_index(gridptr, coords);
  int count = _count_neighbors(gridptr, cmpequal, cmpval, index, gridptr->num_dimensions-1, 1, wrap);
  /* we may have just counted the center itself */
  if ((cmpequal && gridptr->grid_data[index]==cmpval) || 
      (!cmpequal && gridptr->grid_data[index]!=cmpval)) {
    --count;
  }
  return count;
}

int hypergrid_equal_dimensions(hypergrid *gridptr1, hypergrid *gridptr2) {
  int nd = gridptr1->num_dimensions;
  return ( (nd==gridptr2->num_dimensions) &&
           (memcmp(gridptr1->dimension_sizes, gridptr2->dimension_sizes, nd*sizeof(int))==0) );
}


int hypergrid_equal(hypergrid *gridptr1, hypergrid *gridptr2) {
  if (!hypergrid_equal_dimensions(gridptr1, gridptr2)) return 0;
  else {
    int nc = gridptr1->num_cells;
    return ( (nc==gridptr2->num_cells) && 
             (memcmp(gridptr1->grid_data, gridptr2->grid_data, nc*sizeof(gridvalue_t))==0) );
  }
}


void debug_coords(hypergrid *gridptr,int *coords) {
  int i;
  printf("[");
  for(i=0; i<gridptr->num_dimensions; i++) {
    printf("%4d", coords[i]);
  }
  printf("]\n");
}